""" Matplotlib spine represented as a powerpoint line """


#############
## Imports ##
#############

import numpy as np
from scipy.interpolate import interp1d

from .line import Line
from ..templates import LINE
from ..utils.colors import color2hex
from ..utils.strings import random_name
from ..utils.constants import POINTSPERINCH
from ..utils.constants import PIXELSPERPOINT


###########
## Spine ##
###########
class Spine(Line):
    @classmethod
    def from_mpl(cls, mpl_spine):
        """ Create a line starting from a matplotlib Spine object
        
        Args:
            mpl_spine: the matplotlib spine object to convert to a powerpoint line
        
        """
        # Get slidesize from matplotlib figure
        slidesize = (mpl_spine.figure.get_figwidth(), mpl_spine.figure.get_figheight())

        # Get shape
        shape = np.array(mpl_spine.get_extents())

        # Create Line
        spine = cls(
            name="mplspine_" + random_name(5),
            shape=shape,
            lw=mpl_spine._linewidth,
            ec="000000",
            fc=None,
            closed=False,
            slidesize=slidesize,
        )
        return spine



#############
## Imports ##
#############

import os
from ..utils import chdir, get_content


###############
## Constants ##
###############

TEMPLATESPATH = __file__.replace("__init__.pyc", "").replace("__init__.py", "")
PPTXPATH = os.path.join(TEMPLATESPATH, "pptx")


###################
## XML Templates ##
###################

with chdir(TEMPLATESPATH):
    LINE = get_content("line.xml")
    TEXT = get_content("text.xml")
    RECTANGLE = get_content("rectangle.xml")
    IMAGE = get_content("image.xml")



#############
## Imports ##
#############

import os
import shutil
from matplotlib.pyplot import imsave


from .convert import dir2pptx
from .templates import PPTXPATH
from .utils.contextmanagers import chdir
from .utils.contextmanagers import tempdir
from .utils.constants import PIXELSPERINCH


#############
## New PPT ##
#############


def new(filename, xml=None, rels=None, slidesize=(6, 4)):
    """ Creates a new blank powerpoint with a single slide 
    
    Args:
        filename: str: the filename of the new pptx file
        xml: additional xml to insert into the pptx file
        rels: additional rels to insert into the pptx file
        slidesize: the slidesize of the slides in the pptx file
    """
    # check filename
    if filename.endswith(".pptx"):
        filename = filename[:-5]

    # xml should be a string
    if xml is None:
        xml = ""

    # rels should be a list:
    if rels is None:
        rels = []

    # Split path into folder and filename
    splitted_path = os.path.split(filename)
    if splitted_path[:-1] == ("",):
        dir = "."
    else:
        dir = os.path.join(splitted_path[:-1])
    filename = splitted_path[-1]

    # Go to folder
    with chdir(dir):
        # Create a temporary folder to work in
        with tempdir(PPTXPATH) as tempdirname:
            # Change slide size
            presname = os.path.join(tempdirname, "ppt/presentation.xml")
            with open(presname, "r") as file:
                content = file.read()
            content = content.format(
                cx=int(slidesize[0] * PIXELSPERINCH),
                cy=int(slidesize[1] * PIXELSPERINCH),
            )
            with open(presname, "w") as file:
                file.write(content)

            # Insert custom xml
            slidename = os.path.join(tempdirname, "ppt/slides/slide1.xml")
            with open(slidename, "r") as file:
                content = file.read()
            content = content.replace("{objects}", xml)
            with open(slidename, "w") as file:
                file.write(content)

            # Insert custom relationships (for images)
            relsname = os.path.join(tempdirname, "ppt/slides/_rels/slide1.xml.rels")
            with open(relsname, "r") as file:
                content = file.read()
            relationships = ""
            for rel, array, target in rels:
                imsave(os.path.join(tempdirname, "ppt/media", target), array)
                relationships += rel
            content = content.format(relationships=relationships)
            with open(relsname, "w") as file:
                file.write(content)

            # Convert temporary folder to pptx
            dir2pptx(tempdirname, filename)



""" Powerpoint polygon representation """


#############
## Imports ##
#############

import numpy as np
from scipy.interpolate import interp1d

from .line import Line
from ..utils.colors import color2hex
from ..utils.strings import random_name
from ..utils.constants import POINTSPERINCH
from ..utils.constants import PIXELSPERPOINT
from ..utils.mpl import get_plotting_area


#############
## Polygon ##
#############


class Polygon(Line):
    """ A polygon. """

    def __init__(
        self,
        name="Polygon",
        shape=[(0, 0), (0, 100), (50, 0)],
        ec="000000",
        fc="eeeeee",
        slidesize=(6, 4),
    ):
        """ A Polygon is nothing more than a Line which is always closed. 
    
        Args:
            name: the xml name of the polygon
            shape: the shape of the polygon: a list of tuples
            lw: the linewidth to draw the polygon in
            ec: the edgecolor to draw the polygon in
            fc: the facecolor to draw the polygon in
            slidesize=(6,4): the slidesize in which to embed the polygon.
        
        """
        Line.__init__(
            self, name=name, shape=shape, ec=ec, fc=fc, closed=True, slidesize=slidesize
        )

    @classmethod
    def from_mpl(cls, mpl_poly):
        """ Create a polygon starting from a matplotlib Polygon object

        Args:
            mpl_poly: the matplotlib polygon to convert into a ppt polygon
        """

        #TODO: The code below gets repeated a lot over the different shapes.
        #      Create a method in the Object class that extrapolates the
        #      x and y values.

        # Get slidesize from matplotlib figure
        slidesize = (mpl_poly.figure.get_figwidth(), mpl_poly.figure.get_figheight())

        # Get plotting area
        slide_x0, slide_x1, slide_y1, slide_y0 = get_plotting_area(mpl_poly.figure)

        # Translate plot data to locations on slide
        plot_x0, plot_x1 = mpl_poly.axes.get_xlim()
        plot_y0, plot_y1 = mpl_poly.axes.get_ylim()

        mx = (slide_x1 - slide_x0) / (plot_x1 - plot_x0)
        sx = 0.5 * slide_x0
        px = sx / mx

        my = (slide_y1 - slide_y0) / (plot_y1 - plot_y0)
        sy = 0.5 * slide_y0
        py = sy / my

        x, y = mpl_poly._get_xy().T

        x = interp1d(
            [plot_x0, plot_x1], [slide_x0, slide_x1], fill_value="extrapolate"
        )(x)
        y = interp1d(
            [plot_y0, plot_y1], [slide_y0, slide_y1], fill_value="extrapolate"
        )(y)

        # HACK: If an object is partly outside the plotting area, we map the values outside to the
        # margin area (over which the (white?) rectangles of the Canvas will later be drawn)
        x[x < 0.5 * slide_x0] = 0.5 * slide_x0
        x[x > slide_x1 + 0.5 * slide_x0] = slide_x1 + 0.5 * slide_x0
        y[y < 0.5 * slide_y1] = 0.5 * slide_y1
        y[y > slide_y0 + 0.5 * slide_y1] = slide_y0 + 0.5 * slide_y1

        shape = np.stack((x, y), axis=1)

        # If object is completely outside plotting area, then we shouldnt show it at all:
        if ((y < slide_y1) | (y > slide_y0) | (x < slide_x0) | (x > slide_x1)).all():
            return None

        # Create Line
        poly = cls(
            name="mplpolygon_" + random_name(5),
            shape=shape,
            ec=color2hex(mpl_poly._edgecolor),
            fc=None if not mpl_poly.fill else color2hex(mpl_poly._facecolor),
            slidesize=slidesize,
        )

        return poly



""" Python objects representing common powerpoint shapes """

#############
## Imports ##
#############

from .base import Object, Group
from .rectangle import Rectangle
from .line import Line
from .text import Text
from .canvas import Canvas
from .polygon import Polygon
from .image import Image
from .image import Mesh
from .spine import Spine



""" Powerpoint line representations """


#############
## Imports ##
#############

import numpy as np
from scipy.interpolate import interp1d

from .base import Object
from ..templates import LINE
from ..utils.colors import color2hex
from ..utils.strings import random_name
from ..utils.constants import POINTSPERINCH
from ..utils.constants import PIXELSPERPOINT
from ..utils.mpl import get_plotting_area


##########
## Line ##
##########


class Line(Object):
    """ A Line """

    def __init__(
        self,
        name="Line",
        shape=[(0, 0), (0, 100), (50, 0)],
        lw=1,
        ec="000000",
        fc=None,
        closed=False,
        slidesize=(6, 4),
    ):
        """ Create a line in powerpoint

        Args:
            name: the xml name of the line
            shape: the shape of the line: a list of tuples
            lw: the linewidth to draw the line in
            ec: the edgecolor to draw the line in
            fc: the facecolor to draw the line in
            closed: wether to close the line or not (connect the last point with the first point of the line)
            slidesize=(6,4): the slidesize in which to embed the line.
        """
        Object.__init__(self, name=name, slidesize=slidesize)
        self.shape = np.array(shape)
        self.lw = lw
        self.ec = ec
        self.fc = fc
        self.closed = closed
        self._xml = LINE

    @property
    def x(self):
        """ X location of the upper left corner of the shape """
        return np.min(self.shape, axis=0)[0]

    @property
    def y(self):
        """ Y location of the upper left corner of the shape """
        return np.min(self.shape, axis=0)[1]

    @property
    def cx(self):
        """ Total width of the shape """
        return (np.max(self.shape, axis=0) - np.min(self.shape, axis=0))[0]

    @property
    def cy(self):
        """ Total height of the shape """
        return (np.max(self.shape, axis=0) - np.min(self.shape, axis=0))[1]

    def get_adjusted_shape(self):
        """ To draw a resize box around the shape, we need to give the coordinates of the
        shape relative to the the upper left corner of the shape

        Returns:
            shape: list: the shape of the line relative to the upper left corner
        """
        shape = self.shape.copy()
        shape[:, 0] -= self.x
        shape[:, 1] -= self.y
        return shape

    @classmethod
    def from_mpl(cls, mpl_line):
        """ Create a line starting from a matplotlib Line2D object

        Args:
            mpl_line: the matplotlib line to convert into a ppt line
        """

        # TODO: The code below gets repeated a lot over the different shapes.
        #      Create a method in the Object class that extrapolates the
        #      x and y values.

        # Get slidesize from matplotlib figure
        slidesize = (mpl_line.figure.get_figwidth(), mpl_line.figure.get_figheight())

        # Get plotting area
        slide_x0, slide_x1, slide_y1, slide_y0 = get_plotting_area(mpl_line.figure)

        # Translate plot data to locations on slide
        plot_x0, plot_x1 = mpl_line.axes.get_xlim()
        plot_y0, plot_y1 = mpl_line.axes.get_ylim()

        mx = (slide_x1 - slide_x0) / (plot_x1 - plot_x0)
        sx = 0.5 * slide_x0
        px = sx / mx

        my = (slide_y1 - slide_y0) / (plot_y1 - plot_y0)
        sy = 0.5 * slide_y0
        py = sy / my

        x = interp1d(
            [plot_x0, plot_x1], [slide_x0, slide_x1], fill_value="extrapolate"
        )(mpl_line._x)
        y = interp1d(
            [plot_y0, plot_y1], [slide_y0, slide_y1], fill_value="extrapolate"
        )(mpl_line._y)

        # HACK: If an object is partly outside the plotting area, we map the values outside to the
        # margin area (over which the (white?) rectangles of the Canvas will later be drawn)
        x[x < 0.5 * slide_x0] = 0.5 * slide_x0
        x[x > slide_x1 + 0.5 * slide_x0] = slide_x1 + 0.5 * slide_x0
        y[y < 0.5 * slide_y1] = 0.5 * slide_y1
        y[y > slide_y0 + 0.5 * slide_y1] = slide_y0 + 0.5 * slide_y1

        shape = np.stack((x, y), axis=1)

        # If object is completely outside plotting area, then we shouldnt show it at all:
        if ((y < slide_y1) | (y > slide_y0) | (x < slide_x0) | (x > slide_x1)).all():
            return None

        # Create Line
        line = cls(
            name="mplline_" + random_name(5),
            shape=shape,
            lw=mpl_line._linewidth,
            ec=color2hex(mpl_line.get_color()),
            fc=None,
            closed=False,
            slidesize=slidesize,
        )
        return line

    def xml(self):
        """ Get the xml representation of the whole object containing the line 
        
        Returns:
            xml: str: the xml representation of the whole object containing the line
        """
        xml = self._xml.format(
            name=self.name,
            x=int(self.x * PIXELSPERPOINT) + 1,
            y=int(self.y * PIXELSPERPOINT) + 1,
            cx=int(self.cx * PIXELSPERPOINT),
            cy=int(self.cy * PIXELSPERPOINT),
            lw=int(self.lw * PIXELSPERPOINT),
            shapespec=self.shapespec(self.get_adjusted_shape(), self.closed),
            colorspec=self.colorspec(self.ec),
            bgcolorspec=self.colorspec(self.fc),
        )
        return xml

    def shapespec(self, shape, closed):
        """ Get the xml representation of just the line. """
        # check if shape is empty
        if len(shape) == 0:
            return ""

        # first point in shape
        shapespec = '<a:moveTo><a:pt x="{x}" y="{y}"/></a:moveTo>\n'.format(
            x=int(shape[0][0] * PIXELSPERPOINT), y=int(shape[0][1] * PIXELSPERPOINT)
        )

        # other points in shape
        for x, y in shape[1:]:
            shapespec = shapespec + '<a:lnTo><a:pt x="{x}" y="{y}"/></a:lnTo>\n'.format(
                x=int(x * PIXELSPERPOINT), y=int(y * PIXELSPERPOINT)
            )

        # close shape
        if closed:
            shapespec = shapespec + "<a:close/>\n"

        return shapespec



""" `mplppt` seeks to implement an alternative `savefig` function for `matplotlib` figures. 
This `savefig` function saves a `matplotlib` figure with a single axis to a powerpoint 
presentation with a single slide containing the figure. """

__author__ = "Floris Laporte"
__version__ = "0.0.0"

##################
## Unicode Hack ##
## For Python 2 ##
##################

try:  # Python 2
    import sys

    reload(sys)
    sys.setdefaultencoding("utf8")
except NameError:  # Python 3
    pass

################
## Submodules ##
################

from . import shapes
from . import convert
from . import templates
from . import utils


############
## Shapes ##
############

from .shapes import Group
from .shapes import Line
from .shapes import Text
from .shapes import Canvas
from .shapes import Polygon
from .shapes import Rectangle


########################
## The Magic Function ##
########################

from .save import savefig


####################
## New powerpoint ##
####################

from .new import new




""" String and textfile manipulations """


#############
## Imports ##
#############

import string
import random


###############
## Functions ##
###############


def random_name(N=10):
    """ String of N random characters 
    
    Args:
        N: int: the number of random characters

    Returns:
        s: string: the random string with N characters
    """
    return "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(N)
    )


def get_content(filename):
    """ Get content of text file as string 
    
    Args:
        filename: the filename of the textfile to read
    
    Returns:
        s: string: the content of the texfile
    """
    with open(filename, "r") as file:
        content = file.read()
    return content


def parse_xml(filename):
    """ Parse xml file and introduce returns
    
    makes the xml file a bit more readable

    Args:
        filename: str: the filename of the xml file to make more readable
    """
    # check filename
    if not filename.endswith(".xml"):
        filename = filename + ".xml"

    # read content
    with open(filename, "r") as file:
        content = file.read()

    # replace
    content = content.replace("><", ">\n<")

    # overwrite
    with open(filename, "w") as file:
        file.write(content)



""" Filetype conversions """


#############
## Imports ##
#############

import os
import sys
import shutil
import zipfile

from .utils.strings import parse_xml
from .utils.contextmanagers import chdir


#################
## Conversions ##
#################


def dir2pptx(dirname, target=None):
    """ Create ppt file from folder with right ppt structure 
    
    Args:
        dirname: str: name of the directory to convert to a powerpoint file
        target=None: the filename of the resulting pptx file
    """
    # check target
    if target is None:
        target = dirname + ".pptx"
    elif not target.endswith(".pptx"):
        target = target + ".pptx"
    # create zip (pptx) file
    with zipfile.ZipFile(target, "w") as zf:
        # go to folder where zip (pptx) file needs to be created
        with chdir(dirname):
            # add content of folder to zip (pptx) file
            for dir, subdirs, files in os.walk("."):
                zf.write(dir)
                for file in files:
                    zf.write(os.path.join(dir, file))


def dir2zip(dirname, target=None):
    """ Create zip file from folder 
    
    Args:
        dirname: str: name of the directory to convert to a zip file
        target=None: the filename of the resulting zip file.
    """
    # check target
    if target is None:
        target = dirname + ".zip"
    elif not target.endswith(".zip"):
        target = target + ".zip"
    # create zip file
    with zipfile.ZipFile(target, "w") as zf:
        # go to folder where zip file needs to be created
        with chdir(dirname):
            # add content of folder to zip file
            for dir, subdirs, files in os.walk("."):
                zf.write(dir)
                for file in files:
                    zf.write(os.path.join(dir, file))


def pptx2dir(filename, target=None):
    """ Create folder from pptx file
    
    Args:
        filename: str: name of the pptx file to unzip into a directory
        target=None: the filename of the resulting directory
    """
    # check filename:
    filename, ext = os.path.splitext(filename)
    if ext == "":
        ext = ".pptx"

    # check target
    if target is None:
        target = filename

    # unzip
    with zipfile.ZipFile(filename + ext, "r") as zf:
        zf.extractall(target)

    # parse dir for more readable xml files:
    for dir, subdirs, files in os.walk(target):
        for file in files:
            if file.endswith(".xml"):
                parse_xml(os.path.join(dir, file))


def zip2dir(filename, target=None):
    """ Create folder from zip file
    
    Args:
        filename: str: name of the zip file to unzip into a directory
        target=None: the filename of the resulting directory
    """
    # check filename
    filename, ext = os.path.splitext(filename)
    if ext == "":
        ext = ".zip"
    return pptx2dir(filename + ext, target)


def pptx2zip(filename, target=None):
    """ Create zip file from pptx file

    Args:
        filename: str: name of the pptx file to convert to a zip file
        target=None: the filename of the resulting zip file
    """
    # check filename:
    filename, ext = os.path.splitext(filename)
    if ext == "":
        ext = ".pptx"
    filename = filename + ext

    # check target
    if target is None:
        target = filename.replace(".pptx", ".zip")

    # rename
    if os.path.exists(target):
        os.remove(target)
    shutil.copy(filename, target)


def zip2pptx(filename, target=None):
    """ Create pptx file from zip file

    Args:
        filename: str: name of the zip file to convert to a pptx file
        target=None: the filename of the resulting pptx file
    """
    # check filename:
    filename, ext = os.path.splitext(filename)
    if ext == "":
        ext = ".zip"
    filename = filename + ext

    # check target
    if target is None:
        target = filename.replace(".zip", ".pptx")

    # rename
    if os.path.exists(target):
        os.remove(target)
    shutil.copy(filename, target)



""" Powerpoint text representations """


#############
## Imports ##
#############

import numpy as np
from scipy.interpolate import interp1d

from .base import Object
from ..templates import TEXT
from ..templates import RECTANGLE
from ..utils.colors import color2hex
from ..utils.strings import random_name
from ..utils.constants import ALIGNMENTS
from ..utils.constants import POINTSPERINCH
from ..utils.constants import PIXELSPERPOINT
from ..utils.mpl import get_plotting_area


##########
## Text ##
##########


class Text(Object):
    """ A text box """

    def __init__(
        self,
        text="",
        x=0,
        y=0,
        cx=None,
        cy=None,
        size=18,
        font="Arial",
        color="000000",
        ha="c",
        va="c",
        slidesize=(6, 4),
    ):
        """ create a text box in powerpoint

        Args:
            text: str: the text to place in a powerpoint textbox
            x: the x-location of the textbox
            y: the y-location of the textbox
            cx: the x-width of the textbox
            cy: the y-height of the textbox
            size: the fontsize of the text in the textbox
            font: the font of the text in the textbox
            color: the color of the text in the textbox
            ha: the horizontal alignment of the text in the textbox
            va: the vertical alignment of the text in the textbox
            slidesize: the size of the slide to place the textbox into
        """
        Object.__init__(self, name="", slidesize=slidesize)
        self.text = text
        self._x = x
        self._y = y
        self.cx = cx
        self.cy = cy
        self.ha = ALIGNMENTS[ha]
        self.va = ALIGNMENTS[va]
        self.size = size
        self.color = color
        self.font = font
        self._xml = RECTANGLE.replace("</p:sp>", "\n" + TEXT + "\n</p:sp>\n")

    @property
    def x(self):
        """ get the effective x-location of the text adjusted for its horizontal alignment """
        if self.ha == "c":
            return self._x - 0.5 * self.cx
        if self.ha == "r":
            return self._x - self.cx
        return self._x

    @property
    def y(self):
        """ get the effective y-location of the text, adjusted for its vertical alignment """
        if self.va == "c":
            return self._y - 0.5 * self.cy
        if self.va == "b":
            return self._y - self.cy
        return self._y

    @classmethod
    def from_mpl(cls, mpl_text):
        """ Create a text box starting from a matplotlib Text object

        Args:
            mpl_text: the matplotlib text to convert into powerpoint text.
        
        """

        # TODO: The code below gets repeated a lot over the different shapes.
        #       Create a method in the Object class that extrapolates the
        #       x and y values.

        # Get slidesize from matplotlib figure
        slidesize = (mpl_text.figure.get_figwidth(), mpl_text.figure.get_figheight())
        f = cls._mpl_shrink_factor

        # Get plotting area
        slide_x0, slide_x1, slide_y1, slide_y0 = get_plotting_area(mpl_text.figure)

        # Translate text location data to locations on slide
        plot_x0, plot_x1 = mpl_text.axes.get_xlim()
        plot_y0, plot_y1 = mpl_text.axes.get_ylim()

        x = interp1d(
            [plot_x0, plot_x1], [slide_x0, slide_x1], fill_value="extrapolate"
        )(mpl_text._x)
        y = interp1d(
            [plot_y0, plot_y1], [slide_y0, slide_y1], fill_value="extrapolate"
        )(mpl_text._y)

        # HACK: If an object is partly outside the plotting area, we map the values outside to the
        # margin area (over which the (white?) rectangles of the Canvas will later be drawn)
        # TODO: Implement this

        # If object is completely outside plotting area, then we shouldnt show it at all:
        # TODO: Implement this check

        # Get texbox size
        bbox = np.array(
            mpl_text.get_window_extent(renderer=mpl_text.figure.canvas.get_renderer())
        )
        cx, cy = 1.1 * (bbox[1] - bbox[0])

        # Create Textbox
        text = cls(
            text=mpl_text._text,
            x=x,
            y=y,
            cx=abs(cx),
            cy=abs(cy),
            size=mpl_text.get_fontsize(),
            font=mpl_text.get_fontname(),
            color=mpl_text.get_color(),
            ha=mpl_text._horizontalalignment,
            va=mpl_text._verticalalignment,
            slidesize=slidesize,
        )

        return text

    def xml(self):
        """ Get xml representation of the text 
        
        Returns:
            xml: str: the xml representation of the text
        """
        if self.text.replace(" ", "").replace("\n", "") == "":
            return ""  # Return empty string, if no characters are written

        # Get width and height of textbox
        cx = self.cx
        if cx is None:
            cx = self.size * max([len(line) for line in self.text.splitlines()])
        cy = self.cy
        if cy is None:
            cy = self.size + 2

        # Return xml representation
        xml = self._xml.format(
            text=self.text,
            size=int(self.size * 100),
            color=color2hex(self.color),
            font=self.font,
            name=self.text.splitlines()[0],
            x=int(self.x * PIXELSPERPOINT) + 1,
            y=int(self.y * PIXELSPERPOINT) + 1,
            cx=int(cx * PIXELSPERPOINT),
            cy=int(cy * PIXELSPERPOINT),
            lw=int(0.2 * PIXELSPERPOINT),
            colorspec=self.colorspec(None),
            bgcolorspec=self.colorspec(None),
        )
        return xml



""" matplotlib figure tools """


#############
## Imports ##
#############

import os
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from .strings import random_name


###############
## CONSTANTS ##
###############

visualized = {}


###############
## Functions ##
###############


def get_plotting_area(fig):
    """ get area which is visualized by matplotlib

    Args:
        fig: matplotlib figure to find the area for
    
    Returns:
        xmin, xmax, ymin, ymax: the bounds of the matplotlib figure
    """
    global visualized
    if fig not in visualized:
        # HACK: To get info about spine locations, the axis needs to be visualized first.
        # We choose png export:
        fn = random_name() + ".png"
        plt.savefig(fn)
        os.remove(fn)
        visualized[fig] = True
    spines = plt.findobj(fig, mpl.spines.Spine)
    bboxes = [np.array(spine.get_extents()) for spine in spines]
    xmin = np.min([np.min(bbox[:, 0]) for bbox in bboxes])
    xmax = np.max([np.max(bbox[:, 0]) for bbox in bboxes])
    ymin = np.min([np.min(bbox[:, 1]) for bbox in bboxes])
    ymax = np.max([np.max(bbox[:, 1]) for bbox in bboxes])
    return xmin, xmax, ymin, ymax



#############
## Imports ##
#############

import os
import matplotlib as mpl
from matplotlib.pyplot import gcf
from matplotlib.pyplot import findobj

from .shapes import Line
from .shapes import Text
from .shapes import Mesh
from .shapes import Spine
from .shapes import Group
from .shapes import Canvas
from .shapes import Polygon
from .shapes import Rectangle
from .utils.strings import random_name


########################
## The Magic Function ##
########################

# This is the function this repository is all about

def savefig(filename, fig=None, axis=True):
    """ Export a matplotlib figure to a pptx file 
    
    Args:
        filename: str: the filename of the pptx file to save the matplotlib figure as
        fig: the figure to convert to a pptx slide. If None, plt.gcf() will be used to get the most recent figure.
        axis=True: wether to show the axis ticks and labels or not.
    
    Returns:
        group: the mplppt group containing all the objects that were converted from the matplotlib figure.
    """
    # Get figure to save
    if fig is None:
        fig = gcf()

    # Create ppt group
    p = Group(objects=[])

    # Parse mpl objects:
    for obj in findobj(fig):
        # only keep objects that have an axis:
        if obj.axes is not None:
            # convert lines:
            if isinstance(obj, mpl.lines.Line2D):
                p += Line.from_mpl(obj)
            # convert rectangles:
            if isinstance(obj, mpl.patches.Rectangle):
                p += Rectangle.from_mpl(obj)
            # convert polygons
            if isinstance(obj, mpl.patches.Polygon):
                p += Polygon.from_mpl(obj)
            # convert text
            if isinstance(obj, mpl.text.Text):
                p += Text.from_mpl(obj)
            # convert pcolormesh
            if isinstance(obj, mpl.collections.QuadMesh):
                p += Mesh.from_mpl(obj)

    # create a canvas
    # TODO: Create this with less parameters
    canvas = Canvas.from_mpl(fig.axes[0], axis=axis)
    p += canvas

    # save powerpoint group
    p.save(filename)

    # return powerpoint group
    return p


############################
## Other Useful Functions ##
############################

import pickle


def picklefig(filename, fig):
    with open(filename, "wb") as file:
        pickle.dump((fig, fig.canvas), file)


def loadpicklefig(filename):
    with open(filename, "rb") as file:
        try:
            fig, canvas = pickle.load(file, encoding="latin1")
        except TypeError:
            fig, canvas = pickle.load(file)
    fig.canvas = canvas
    return fig



""" useful context managers for mplppt """

#############
## Imports ##
#############

import os
import shutil
from contextlib import contextmanager

from .strings import random_name


######################
## Context Managers ##
######################


@contextmanager
def chdir(path):
    """ Handy working directory changing context manager that returns to 
    original folder if something goes wrong 
    
    Args:
        path: str: path of folder to cd into
    """
    try:
        old_dir = os.getcwd()
        os.chdir(path)
        yield
    finally:
        os.chdir(old_dir)


@contextmanager
def tempdir(source=None):
    """ Creates a temporary folder (empty if no source provided). 
    The folder gets automatically removed if something goes wrong.
    
    Args:
        source: the source folder to copy into the temporary folder
    """
    try:
        dirname = random_name()
        if source is None:
            os.mkdir(dirname)
        else:
            shutil.copytree(source, dirname)
        yield dirname
    finally:
        shutil.rmtree(dirname)



""" Powerpoint rectangle representation """


#############
## Imports ##
#############

from .base import Object
from ..templates import RECTANGLE
from ..utils.colors import color2hex
from ..utils.strings import random_name
from ..utils.constants import POINTSPERINCH
from ..utils.constants import PIXELSPERPOINT
from ..utils.mpl import get_plotting_area


###############
## Rectangle ##
###############


class Rectangle(Object):
    """ A rectangle """

    def __init__(
        self,
        name="Rect",
        x=0,
        y=0,
        cx=100,
        cy=50,
        lw=1,
        ec="000000",
        fc="eeeeee",
        slidesize=(6, 4),
    ):
        """ Create a powerpoint rectangle

        Args:
            name: the xml-name for the rectangle
            x=0: the x-location of the rectangle
            y=0: the y-location of the rectangle
            cx=0: the width of the rectangle
            cy=0: the height of the rectangle
            lw: the linewidth to draw the rectangle in
            ec: the edgecolor to draw the rectangle in
            fc: the facecolor to draw the rectangle in
            slidesize=(6,8): the slidesize to put the rectangle in
        """
        Object.__init__(self, name=name, slidesize=slidesize)
        self.x = x
        self.y = y
        self.cx = cx
        self.cy = cy
        self.lw = lw
        self.ec = ec
        self.fc = fc
        self._xml = RECTANGLE

    @classmethod
    def from_mpl(cls, mpl_rect):
        """ Create a rectangle starting from a matplotlib Rectangle object

        Args:
            mpl_rect: the matplotlib rectangle to convert to a powerpoint rectangle
        """

        # TODO: The code below gets repeated a lot over the different shapes.
        #       Create a method in the Object class that extrapolates the
        #       x and y values.

        # Get slidesize from matplotlib figure
        slidesize = (mpl_rect.figure.get_figwidth(), mpl_rect.figure.get_figheight())

        # Get plotting area
        slide_x0, slide_x1, slide_y1, slide_y0 = get_plotting_area(mpl_rect.figure)

        rx, ry = mpl_rect.xy

        # Translate plot data to locations on slide
        plot_x0, plot_x1 = mpl_rect.axes.get_xlim()
        mx = (slide_x0 - slide_x1) / (plot_x0 - plot_x1)
        x = mx * (rx - plot_x0) + slide_x0
        cx = mx * mpl_rect._width
        if cx < 0:
            x += cx
            cx *= -1

        plot_y0, plot_y1 = mpl_rect.axes.get_ylim()
        my = (slide_y0 - slide_y1) / (plot_y0 - plot_y1)
        y = my * (ry - plot_y0) + slide_y0
        cy = my * mpl_rect._height
        if cy < 0:
            y += cy
            cy *= -1

        # HACK: If an object is partly outside the plotting area, we map the values outside to the
        # margin area (over which the (white?) rectangles of the Canvas will later be drawn)
        # TODO: Implement this

        # If object is completely outside plotting area, then we shouldnt show it at all:
        # TODO: Implement this check

        # Create Rectangle
        rect = cls(
            name="mplrect_" + random_name(5),
            x=x,
            y=y,
            cx=cx,
            cy=cy,
            lw=mpl_rect._linewidth,
            ec=color2hex(mpl_rect._edgecolor),
            fc=color2hex(mpl_rect._facecolor),
            slidesize=slidesize,
        )
        return rect

    def xml(self):
        """ Get xml representation of the rectangle """
        xml = self._xml.format(
            name=self.name,
            x=int(self.x * PIXELSPERPOINT) + 1,
            y=int(self.y * PIXELSPERPOINT) + 1,
            cx=int(self.cx * PIXELSPERPOINT),
            cy=int(self.cy * PIXELSPERPOINT),
            lw=int(self.lw * PIXELSPERPOINT),
            colorspec=self.colorspec(self.ec),
            bgcolorspec=self.colorspec(self.fc),
        )
        return xml



""" color conversions """

#############
## Imports ##
#############

from .constants import MPLCOLORS

#####################
## Color Functions ##
#####################


def rgb2hex(tup, max=1):
    """ convert an rgb tuple into a hex color string representation

    Args:
        tup: tuple: tuple of length 3 representing the color
        max=1: the maximum value for the values in the tuple.
            This will be used to scale the tuple values as integers between 0 and 255.
    
    Returns:
        color: str: hex color string representation (format "aaaaaa" - no "#")
    """
    tup = [int(c * 255 / max) for c in tup]
    color = ""
    for c in tup:
        if c < 16:
            c = "0" + hex(c)[-1:]
        else:
            c = hex(c)[-2:]
        color += c
    return color.upper()


def rgba2hex(tup, max=1, keep_alpha=True):
    """ convert an rgba tuple into a hex color string representation and a transparency value (alpha value)

    Args:
        tup: tuple: tuple of length 4 representing the color and a transparency
        max=1: the maximum value for the values in the tuple.
            This will be used to scale the tuple values as integers between 0 and 255.
        keep_alpha=True: wether to keep the transparency or not.
    
    Returns:
        color: tuple: (hex color representation, alpha)
     OR color: str: hex color representation [if keep_alpha=False]
    """
    alpha = tup[-1]
    color = rgb2hex(tup[:-1])
    if keep_alpha:
        return color, alpha
    else:
        return color


def color2hex(color):
    """ Convert any kind of color representation to a hex string color representation 
    
    Args:
        color: any kind of color representation (hex string, tuple)
    
    Returns:
        color: hex string representation for the color (format "aaaaaa" - no "#")
    """
    if isinstance(color, str):
        if len(color) == 6:
            return color
        elif len(color) == 7 and color[0] == "#":
            return color[1:]
        elif color in MPLCOLORS:
            return MPLCOLORS[color]
    else:
        if isinstance(color, int):
            return rgb2hex((color, color, color), max=255)
        if isinstance(color, float):
            return rgb2hex((color, color, color))
        if len(color) == 4:
            return rgba2hex(color)
        if len(color) == 3:
            return rgb2hex(color)
        if len(color) == 2:
            return color
    raise ValueError("invalid color " + str(color))



""" Matplotlib canvas represented as a collection of powerpoint shapes """

#############
## Imports ##
#############

import numpy as np
from scipy.interpolate import interp1d

from .line import Line
from .text import Text
from .base import Group
from .rectangle import Rectangle
from ..utils.constants import POINTSPERINCH
from ..utils.mpl import get_plotting_area
from ..utils.colors import color2hex


############
## Canvas ##
############


class Canvas(Group):
    """ Draw a canvas around a group of objects """
    def __init__(
        self, x, y, cx, cy, lw=0.8, ec="000000", fc="ffffff", slidesize=(6, 4)
    ):
        """ A canvas draws a rectangle around the plot with linewidth lw and edgecolor ec.
        It also hides the parts of the plot that are outside the drawing region
        by inserting 4 rectangles with facecolor fc.

        Args:
            x: the x-location of the canvas
            y: the y-location of the canvas
            cx: the x-width of the canvas
            cy: the y-height of the canvas
            lw: the linewidth to draw the canvas in
            ec:  the edgecolor to draw the canvas in
            fc: the facecolor to draw the area outside the canvas in
            slidesize: the size of the slide to draw the canvas in

        NOTE: This object is a bit of a hack, since the canvas should in
        principle be translated automatically from the mpl_figure. Right
        now, this translation is only partly implemented. (This is why
        .from_mpl accepts some keyword arguments.)
        """

        # Rectangle around the plot
        self.rect = Rectangle(
            name="Canvas",
            x=x * POINTSPERINCH,
            y=y * POINTSPERINCH,
            cx=cx * POINTSPERINCH,
            cy=cy * POINTSPERINCH,
            lw=lw,
            ec=ec,
            fc=None,
            slidesize=slidesize,
        )

        # Rectangles hiding the parts outside the drawing region:
        self.left = Rectangle(
            name="Canvas_left",
            x=0,
            y=0,
            cx=x * POINTSPERINCH,
            cy=slidesize[1] * POINTSPERINCH,
            lw=0.8,
            ec=None,
            fc=fc,
        )
        self.right = Rectangle(
            name="Canvas_right",
            x=(x + cx) * POINTSPERINCH,
            y=0,
            cx=x * POINTSPERINCH,
            cy=slidesize[1] * POINTSPERINCH,
            lw=0.8,
            ec=None,
            fc=fc,
        )
        self.top = Rectangle(
            name="Canvas_top",
            x=0,
            y=0,
            cx=slidesize[0] * POINTSPERINCH,
            cy=y * POINTSPERINCH,
            lw=0.8,
            ec=None,
            fc=fc,
        )
        self.bottom = Rectangle(
            name="Canvas_bottom",
            x=0,
            y=(y + cy) * POINTSPERINCH,
            cx=slidesize[0] * POINTSPERINCH,
            cy=y * POINTSPERINCH,
            lw=0.8,
            ec=None,
            fc=fc,
        )

        # Finalize initialization
        Group.__init__(
            self,
            name="Canvas",
            objects=[self.left, self.right, self.top, self.bottom, self.rect],
        )

    @classmethod
    def from_mpl(cls, mpl_ax, lw=0.8, ec="000000", fc="ffffff", axis=True):
        """ Create a canvas starting from a matplotlib axis
        
        Args:
            mpl_ax: the matplotlib axis object to draw the canvas from
            lw: the linewdith to draw the canvas in
            ec: the edgecolor to draw the canvas in
            fc: the facecolor to draw the canvas in
            axis=True: wether to draw the axis ticks and labels.
         
        """

        # Get slidesize from matplotlib figure
        slidesize = (mpl_ax.figure.get_figwidth(), mpl_ax.figure.get_figheight())

        # Get plotting area
        slide_x0, slide_x1, slide_y1, slide_y0 = get_plotting_area(mpl_ax.figure)

        x = min(slide_x0, slide_x1) / POINTSPERINCH
        y = min(slide_y0, slide_y1) / POINTSPERINCH
        cx = abs(slide_x0 - slide_x1) / POINTSPERINCH
        cy = abs(slide_y0 - slide_y1) / POINTSPERINCH

        # Initialize Canvas
        canvas = cls(
            x=x,
            y=y,
            cx=cx,
            cy=cy,
            lw=lw,
            ec=ec if axis else color2hex((0.0, 0.0, 0.0, 0.0)),
            fc=fc,
            slidesize=slidesize,
        )

        # Add ticks (numbers) to side of plot
        if axis:
            ylim = mpl_ax.get_ylim()
            for mpl_text in mpl_ax.xaxis.get_ticklabels():
                # HACK: I havent found a way to copy the axis.
                # We store the old values and put them back in
                # The figure to not change the original matplotlib axis
                old_axes = mpl_text.axes
                old_text = mpl_text._text
                old_y = mpl_text._y

                mpl_text.axes = mpl_ax
                mpl_text._text = mpl_text._text.replace(
                    "\u2212", "-"
                )  #'\u2212 yields errors while writing to file
                mpl_text._y = ylim[0] - 0.01 * (ylim[1] - ylim[0])
                canvas = canvas + Text.from_mpl(mpl_text)

                mpl_text.axes = old_axes
                mpl_text._text = old_text
                mpl_text._y = old_y

            xlim = mpl_ax.get_xlim()
            for mpl_text in mpl_ax.yaxis.get_ticklabels():
                # HACK: I havent found a way to copy the axis.
                # We store the old values and put them back in
                # The figure to not change the original matplotlib axis
                old_axes = mpl_text.axes
                old_text = mpl_text._text
                old_x = mpl_text._x

                mpl_text.axes = mpl_ax
                mpl_text._text = mpl_text._text.replace(
                    "\u2212", "-"
                )  #'\u2212 yields errors while writing to file
                mpl_text._x = xlim[0] - 0.01 * (xlim[1] - xlim[0])
                canvas = canvas + Text.from_mpl(mpl_text)

                mpl_text.axes = old_axes
                mpl_text._text = old_text
                mpl_text._x = old_x

        # Return canvas
        return canvas



""" Powerpoint image representations """

#############
## Imports ##
#############

import os
import numpy as np
from matplotlib.pyplot import imread, imsave
from scipy.interpolate import interp2d

from .base import Object
from ..templates import IMAGE
from ..utils.strings import random_name
from ..utils.contextmanagers import chdir
from ..utils.constants import PIXELSPERPOINT
from ..utils.constants import POINTSPERINCH
from ..utils.mpl import get_plotting_area


###############
## Rectangle ##
###############


class Image(Object):
    """ A Powerpoint Image """

    def __init__(self, source, name="", x=0, y=0, cx=None, cy=None, slidesize=(6, 4)):
        """ Create a powerpoint image

        Args:
            source: the filename of the image to add to the powerpoint slide
            name: the xml-name for the image
            x=0: the x-location of the image
            y=0: the y-location of the image
            cx=0: the width of the image
            cy=0: the height of the image
            slidesize=(6,8): the slidesize to put the image in
        """
        name = ".".join(source.split(".")[:-1]) if name == "" else name
        Object.__init__(self, name=name, slidesize=slidesize)
        self.array = imread(source)
        self.id = random_name(5)
        self.source = source
        self.target = self.name + "_" + self.id + "." + source.split(".")[-1]
        self.x = x
        self.y = y

        self.cx = 10000 * self.array.shape[1] / PIXELSPERPOINT if cx is None else cx
        self.cy = 10000 * self.array.shape[0] / PIXELSPERPOINT if cy is None else cy
        self._xml = IMAGE

    def rels(self):
        """ The relationship representation of the image 
        
        This relationship information contains the source location of the image, 
        as well as the xml schema used to visualize it.

        Returns:
            rels: list: the list of relationships for the image.
        
        """
        rels = '<Relationship Id="{id}" Type="http://schemas.openxmlformats.org/officeDocument/2006/relationships/image" Target="../media/{target}"/>'
        rels = rels.format(id=self.id, target=self.target)
        return [(rels, self.array, self.target)]

    def xml(self):
        """ Get xml representation of the image 
        
        Returns:
            xml: the xml representation for the image
        """
        xml = self._xml.format(
            name=self.name,
            x=int(self.x * PIXELSPERPOINT) + 1,
            y=int(self.y * PIXELSPERPOINT) + 1,
            cx=int(self.cx * PIXELSPERPOINT),
            cy=int(self.cy * PIXELSPERPOINT),
            id=self.id,
        )
        return xml


class Mesh(Image):
    """ Matplotlib QuadMesh (plt.pcolormesh) representated as a powerpoint image """

    @classmethod
    def from_mpl(cls, mpl_mesh):
        """ create a Mesh from a matplotlib QuadMesh object

        Args:
            mpl_mesh: the matplotlib QuadMesh object to represent as a powerpoint image
        """
        xlim = mpl_mesh.axes.get_xlim()
        ylim = mpl_mesh.axes.get_ylim()
        X = mpl_mesh._coordinates[:-1, :-1, 0]
        Y = mpl_mesh._coordinates[:-1, :-1, 1]
        xmin, xmax = max(np.min(X), min(xlim)), min(np.max(X), max(xlim))
        ymin, ymax = max(np.min(Y), min(ylim)), min(np.max(Y), max(ylim))

        Z = mpl_mesh._A.reshape(mpl_mesh._meshHeight, mpl_mesh._meshWidth).data

        # Get plotting area
        slide_x0, slide_x1, slide_y1, slide_y0 = get_plotting_area(mpl_mesh.figure)

        # Get slidesize from matplotlib figure
        slidesize = (mpl_mesh.figure.get_figwidth(), mpl_mesh.figure.get_figheight())

        # Translate plot data to locations on slide
        plot_x0, plot_x1 = mpl_mesh.axes.get_xlim()
        mx = (slide_x0 - slide_x1) / (plot_x0 - plot_x1)
        x = mx * (xmin - plot_x0) + slide_x0
        cx = mx * (xmax - xmin)
        if cx < 0:
            x += cx
            cx *= -1

        plot_y0, plot_y1 = mpl_mesh.axes.get_ylim()
        my = (slide_y0 - slide_y1) / (plot_y0 - plot_y1)
        y = my * (ymin - plot_y0) + slide_y0
        cy = my * (ymax - ymin)
        if cy < 0:
            y += cy
            cy *= -1

        id = random_name(5)
        _x, _y = np.where(((X > xmin) & (X < xmax) & (Y > ymin) & (Y < ymax)))
        imsave(id + ".png", Z[max(_x) : min(_x) : -1, min(_y) : max(_y)])

        mesh = cls(
            source=id + ".png", name=id, x=x, y=y, cx=cx, cy=cy, slidesize=slidesize
        )
        os.remove(id + ".png")
        mesh.id = id
        mesh.target = id + ".png"
        return mesh



import mplppt
import setuptools

description = """ Convert a matplotlib figure to a powerpoint slide """

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="mplppt",
    version=mplppt.__version__,
    author=mplppt.__author__,
    author_email="floris.laporte@gmail.com",
    description=description.replace("\n", " "),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/flaport/mplppt",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2.7",
        "Development Status :: 3 - Alpha",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],
)



""" Base objects for powerpoint """

#############
## Imports ##
#############

from ..new import new


###################
## Object Parent ##
###################


class Object(object):
    """ An abstract powerpoint object
    
    a powerpoint object can be any base object or a group of other objects.
    """

    # scaling factor for matplotlib figures
    # if this factor is 1, the whole slide will be covered by the plot
    _mpl_shrink_factor = 0.9

    def __init__(self, name="", slidesize=(6, 4)):
        """ powerpoint object initialization

        Args:
            name: str="": the name of the object to embed in the powerpoint presentation
            slidesize=(6,4): the size of the slide the object is embedded in.        
        """

        # TODO: Create a slide object to embed all the slide properties in.

        self.name = name
        self._xml = ""
        self.slidesize = slidesize

    def rels(self):
        """ Get relationship representation of current object 
        
        Returns:
            rels: list: the list of relationships to other objects.
        """
        return []

    def xml(self):
        """ Get xml representation of current object 
        
        Returns:
            xml: str: the xml representation of this object
        """
        return self._xml

    def save(self, filename):
        """ Save current object as powerpoint presentation 
        
        Args:
            filename: str: the filename to save this object under.
        """
        new(filename, xml=self.xml(), rels=self.rels(), slidesize=self.slidesize)

    def colorspec(self, color):
        """ Xml representation of the color of the object. 
        
        Args:
            color: None|str|tuple: color can be a hexstring of format 'aaaaaa', or a tuple consisting of a
                hexstring and an alpha value between 0 and 1. If `None` a `noFill` xml tag will be returned.

        Returns:
            xml: the xml representation for the color of the object.
        """
        if color is None:
            return "<a:noFill/>"
        else:
            if not isinstance(color, str):
                color, alpha = color
                alpha = str(int(alpha * 100000))
            else:
                alpha = "100000"
            return (
                '<a:solidFill><a:srgbClr val="'
                + color
                + '"><a:alpha val="'
                + alpha
                + '"/></a:srgbClr></a:solidFill>'
            )

    def __add__(self, other):
        """ Objects can be added together to form a Group of objects 
        
        Args:
            other: Object: the object to add this object together with into a group

        Returns:
            group: Group: a group of two objects containing this object and the other object.
        """
        return Group(objects=[self, other], slidesize=self.slidesize)

    def __radd__(self, other):
        """ Objects can be added together to form a Group of objects

        [same behavior as __add__]
        
        Args:
            other: Object: the object to add this object together with into a group

        Returns:
            group: Group: a group of two objects containing this object and the other object.
        """
        return self + other


##################
## Object Group ##
##################


class Group(Object):
    """ A Group consists out of a collection of smaller objects (other groups are allowed).

    NOTE: A mplppt Group does NOT correspond to a Powerpoint Group. It is just a group in
    python and acts as an abstract object.
    """

    # TODO: Implement the expected behavior such that an mplppt group corresponds to a powerpoint group.

    def __init__(self, name="ppt", objects=[], slidesize=(6, 4)):
        Object.__init__(self, name=name, slidesize=slidesize)
        self.objects = objects

    def rels(self):
        """ Get relationship representation of current object 
        
        Returns:
            rels: list: the list of relationships to other objects.
        """
        rels = []
        for obj in self.objects:
            rels += obj.rels()
        return rels

    def xml(self):
        """ Get xml representation of current object 
        
        Returns:
            xml: str: the xml representation of this object
        """
        xml = ""
        for obj in self.objects:
            xml += "\n" + obj.xml() + "\n"
        return xml

    def __add__(self, other):
        """ Objects can be added together to form a Group of objects 
        
        Args:
            other: Object: the object to add this object together with into a group

        Returns:
            group: Group: a group of two objects containing this object and the other object.
        """
        # TODO: remove this method in favor of the more generic __add__ method of Object.
        if other is None: # if nothing is added to the group, one should return the original group
            return self
        # prefer this slidesize if it's defined, otherwise take over the slidesize of the other group
        slidesize = other.slidesize if len(self.objects) == 0 else self.slidesize
        if hasattr(other, "objects"): # if the other object is a group, merge the two groups
            return Group(objects=self.objects + other.objects, slidesize=slidesize)
        else: # if the other object is an object, add it to the objects
            return Group(objects=self.objects + [other], slidesize=slidesize)

    def __iadd__(self, other):
        """ Objects can be added into the group 
        
        Args:
            other: Object: the object to add into the group
        
        Returns:
            group: Group: the resulting group after the addition of the other object
        """
        if other is not None:
            self.slidesize = (
                other.slidesize if len(self.objects) == 0 else self.slidesize
            )
            if hasattr(other, "objects"):
                self.objects += other.objects
            else:
                self.objects += [other]
        return self

    def __sub__(self, other):
        """ Objects can be removed from the group 
        
        Args:
            other: Object: the object to remove from the group
        
        Returns:
            group: Group: the resulting group after the removal of the object
        """
        if other is None:
            return self
        if other not in self.objects:
            raise ValueError("Group does not contain object %s" % object.name)
        return Group(
            objects=[obj for obj in self.objects if obj is not other],
            slidesize=self.slidesize,
        )

    def __isub__(self, other):
        """ Objects can be removed from the group 
        
        Args:
            other: Object: the object to remove from the group
        
        Returns:
            group: Group: the resulting group after the removal of the object
        """
        if other is not None:
            i = self.objects.index(other)
            del self.objects[i]
        return self



""" Some basic undocumented CLI behavior. Have fun! """

if __name__ == "__main__":
    import os
    import sys

    if sys.argv[1] == "new":
        from .new import new

        fn = "new.pptx" if len(sys.argv) < 3 else sys.argv[2]
        w = 13.333 if "-w" not in sys.argv else sys.argv[sys.argv.index("-w") + 1]
        h = 7.5 if "-h" not in sys.argv else sys.argv[sys.argv.index("-h") + 1]
        new(fn, slidesize=(w, h))

    if sys.argv[1] == "convert" or sys.argv[1] == "c":
        from .convert import dir2pptx, dir2zip, pptx2dir, zip2dir, zip2pptx, pptx2zip

        fn, ext = os.path.splitext(sys.argv[2])

        if ext == "":
            if os.path.isdir(fn):
                frm = "dir"
            elif os.path.exists(fn + ".pptx"):
                frm = "pptx"
            elif os.path.exists(fn + ".zip"):
                frm = "zip"
        else:
            frm = ext[1:]

        if "-to" not in sys.argv:
            if frm == "pptx" or frm == "zip":
                to = "dir"
            elif frm == "dir":
                to = "pptx"
        else:
            to = sys.argv[sys.argv.index("-to") + 1]
            if to == "ppt":
                to = "pptx"

        func = {
            "dirpptx": dir2pptx,
            "dirzip": dir2zip,
            "pptxdir": pptx2dir,
            "zipdir": zip2dir,
            "zippptx": zip2pptx,
            "pptxzip": pptx2zip,
        }[frm + to]

        if frm == "dir":
            func(fn)
        else:
            func(fn + "." + frm)
